from typing import Optional
import logging


from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

logger = logging.getLogger('factro_client')

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Contact:
    id: str
    company_id: str
    description: Optional[str] = ''
    city: Optional[str] = ''
    email_address: Optional[str] = ''
    phone: Optional[str] = ''
    street: Optional[str] = ''
    zip_code: Optional[str] = ''
    first_name: Optional[str] = ''
    last_name: Optional[str] = ''
    mobile_phone: Optional[str] = ''
    salutation: Optional[str] = ''

class ContactHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(ContactHandler, self).__init__(url, api_key)

    def get_contact_by_id(self, contact_id) -> Contact:
        path = f"/api/core/contacts/{contact_id}"

        json_data = self.do_get(path)

        logger.debug(f"ContactHandler.get_contact_by_id({contact_id} -> {json_data}")

        contact = Contact.schema().load(json_data)

        return contact


from datetime import datetime, timedelta
import logging
import os
from typing import List, Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase
from cachetools import cached, TTLCache

from tzlocal import get_localzone


from ..core.client import Client as CoreClient
from ..core.client import WorkRecord, Company, Contact, Project, Package, Task, User

logger = logging.getLogger('factro_client')


CACHE_MAX_SIZE = os.environ.get('FACTRO_CLIENT_CACHE_MAX_SIZE', 100)
CACHE_TTL = os.environ.get('FACTRO_CLIENT_CACHE_TTL', 2 * 60 * 60)  # 2 hours

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ActivityRecording:
    start_date: str
    end_date: str
    title: str
    is_billable: bool
    hours: float
    description: str
    driving_distance_km: Optional[float]
    task_number: str
    company_name: str
    company_shortname: str
    project_name: str
    project_number: str
    project_tags: Optional[str]
    package_id: Optional[str]
    package_number: Optional[str]
    package_title: Optional[str]
    package_custom_fields: Optional[dict]
    email: str
    project_role: str
    is_billable_raw: bool
    billing_cycle: str
    offer_number: str
    order_number: str
    firstname: str
    lastname: str
    cost_center: Optional[str]
    department: Optional[str]


class Client:

    def __init__(self, url: str, api_key: str):
        self._core_client = CoreClient(url, api_key)

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_company_by_id(self, company_id: str) -> Company:
        company = None

        if company_id is not None:
            company = self._core_client.get_company_by_id(company_id)

        return company

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_contact_by_id(self, contact_id: str) -> Contact:
        contact = None

        if contact_id is not None:
            contact = self._core_client.get_contact_by_id(contact_id)

        return contact

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_project_by_id(self, project_id: str) -> Project:
        project = self._core_client.get_project_by_id(project_id)
        return project

    @cached(cache=TTLCache(maxsize=100, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_package_by_id(self, package_id: str) -> Package:
        if package_id is None:
            return None

        package = self._core_client.get_package_by_id(package_id)
        return package

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_task_by_id(self, task_id: str) -> Task:
        task = self._core_client.get_task_by_id(task_id)
        return task

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_user_by_id(self, user_id: str) -> User:
        user = self._core_client.get_user_by_id(user_id)
        return user

    def query_controlling_data(self, from_date: str, to_date: str) -> List[ActivityRecording]:

        tz = get_localzone()

        def is_billable(workrecord: WorkRecord, package: Package, project: Project) -> bool:
            billable = package.custom_fields.get('abrechenbar', workrecord.is_billable) if package is not None else workrecord.is_billable

            if billable:
                if project is not None and "abrechnungsart" in project.custom_fields.keys():
                    billable = project.custom_fields.get('abrechnungsart') != 'keine_abrechnung'

            return billable

        def make_datetime(date_string: str) -> datetime:
            d = datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S')

            return tz.localize(d).isoformat()

        def mapping(workrecord: WorkRecord, user: User, task: Task, package: Package, project: Project, company: Company) -> ActivityRecording:
            activity_recording = ActivityRecording(
                start_date = make_datetime(f"{workrecord.start_date}T{workrecord.start_time}"),
                end_date = make_datetime(f"{workrecord.end_date}T{workrecord.end_time}"),
                title = workrecord.title,
                #is_billable = package.custom_fields.get('abrechenbar', workrecord.is_billable) if package is not None else workrecord.is_billable,
                is_billable = is_billable(workrecord, package, project),
                hours = workrecord.minutes_worked / 60,
                description = workrecord.description,
                driving_distance_km = 0,
                task_number = task.number,
                company_name = company.name if company is not None else '',
                company_shortname = company.short_name if company is not None else '',
                project_name = project.title,
                project_number = project.number,
                project_tags = [], #project.tags,
                package_id = package.id if package is not None else '',
                package_number = package.number if package is not None else '',
                package_title = package.title if package is not None else '',
                package_custom_fields = package.custom_fields if package is not None else {},
                email = user.email_address,
                project_role = task.custom_fields.get('rolle', ''),
                is_billable_raw = workrecord.is_billable,
                billing_cycle = project.custom_fields.get('abrechnungsart', ''),
                offer_number = project.custom_fields.get('angebotsnummer', ''),
                order_number = project.custom_fields.get('bestellnummer', ''),
                firstname = user.first_name,
                lastname = user.last_name,
                cost_center = '', #Optional[str],
                department = '', #Optional[str],
            )

            return activity_recording

        logger.debug(f"query controlling data from {from_date} to {to_date}")
        activity_recordings = []

        workrecords = self._core_client.get_work_records(from_date, to_date)


        for workrecord in workrecords:
            user = self._query_user_by_id(workrecord.work_employee_id)
            task = self._query_task_by_id(workrecord.booked_on_reference_id)
            package = self._query_package_by_id(task.parent_package_id)
            project = self._query_project_by_id(task.project_id)
            company = self._query_company_by_id(project.company_id) if project is not None else None

            activity_recording = mapping(workrecord, user, task, package, project, company)

            activity_recordings.append(activity_recording)

        return activity_recordings



from typing import Optional
import logging


from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

logger = logging.getLogger('factro_client')

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Task:
    id: str
    mandant_id: str
    number: int
    change_date: str
    creator_id: str
    title: str
    officer_id: str
    planned_effort: float
    project_id: str
    realized_effort: float
    remaining_effort: float
    executor_id: str
    is_milestone: bool
    task_state: str
    custom_fields: Optional[dict]
    parent_package_id: str = None
    company_contact_id: Optional[str] = ''
    description: Optional[str] = ''
    end_date: Optional[str] = ''
    start_date: Optional[str] = ''
    company_id: Optional[str] = ''
    paused_until: Optional[str] = ''
    task_priority: Optional[int] = 0


class TaskHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(TaskHandler, self).__init__(url, api_key)

    def get_task_by_id(self, task_id) -> Task:
        path = f"/api/core/tasks/{task_id}"

        json_data = self.do_get(path)

        logger.debug(f"TaskHandler.get_task_by_id({task_id} -> {json_data}")
        try:
            task = Task.schema().load(json_data)
        except Exception as e:
            logger.error(f"TaskHandler.get_task_by_id({task_id} -> {json_data})")
            raise e

        return task


from typing import List, Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class WorkRecord:
    creator_employee_id: str
    description: str
    end_date: str
    end_time: str
    id: str
    is_billable: bool
    is_billed: bool
    mandant_id: str
    minutes_worked: float
    start_date: str
    start_time: str
    title: str
    work_employee_id: str
    booked_on_reference_id: Optional[str] = ''
    external_booking_details: Optional[str] = ''
    travelled_distance_km: Optional[float] = 0.0
    internal_booking_details: Optional[str] = ''
    created_in_context_of_reference_id: Optional[str] = ''


class WorkRecordHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(WorkRecordHandler, self).__init__(url, api_key)

    def get_work_records(self, from_date:str, to_date:str) -> List[WorkRecord]:
        query = f"?startDate={from_date}&endDate={to_date}"
        path = f"/api/core/work-records{query}"

        json_data = self.do_get(path)

        work_records = WorkRecord.schema().load(json_data, many=True)

        return work_records


from typing import Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class User:
    id: str
    mandant_id: str
    account_id: str
    anonymized: bool
    employee_number: int
    is_active: bool
    security_group: str
    user_name: str
    city: Optional[str] = ''
    email_address: Optional[str] = ''
    street: Optional[str] = ''
    zip_code: Optional[str] = ''
    first_name: Optional[str] = ''
    last_name: Optional[str] = ''
    salutation: Optional[str] = ''

class UserHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(UserHandler, self).__init__(url, api_key)

    def get_user_by_id(self, user_id) -> User:
        path = f"/api/core/users/{user_id}"

        json_data = self.do_get(path)

        user = User.schema().load(json_data)

        return user


from typing import Optional, List
import logging


from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

logger = logging.getLogger('factro_client')

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Company:
    id: str
    description: Optional[str] = ''
    city: Optional[str] = ''
    company_number: Optional[int] = ''
    email_address: Optional[str] = ''
    name: Optional[str] = ''
    phone: Optional[str] = ''
    short_name: Optional[str] = ''
    street: Optional[str] = ''
    website: Optional[str] = ''
    zip_code: Optional[str] = ''
    customer_id: Optional[str] = ''
    contact_ids: Optional[List[str]] = ''


class CompanyHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(CompanyHandler, self).__init__(url, api_key)

    def get_company_by_id(self, company_id) -> Company:
        path = f"/api/core/companies/{company_id}"

        json_data = self.do_get(path)

        logger.debug(f"CompanyHandler.get_company_by_id({company_id} -> {json_data}")

        company = Company.schema().load(json_data)

        return company



class UnauthorizedException(Exception):
    
    MESSAGE = """
        The request failed cause an `401` status code raised.

        Please check the used api_key, maybe it is expired or missing.
    """

    def __init__(self):
        super(UnauthorizedException, self).__init__(UnauthorizedException.MESSAGE)



from typing import List

from .helpers.work_record_handler import WorkRecordHandler, WorkRecord
from .helpers.user_handler import UserHandler, User
from .helpers.task_handler import TaskHandler, Task
from .helpers.package_handler import PackageHandler, Package
from .helpers.project_handler import ProjectHandler, Project, ProjectStructure
from .helpers.contact_handler import ContactHandler, Contact
from .helpers.company_handler import CompanyHandler, Company


class Client(object):

    def __init__(self, url:str, api_key:str):
        self._url = url
        self._api_key = api_key

    def get_work_records(self, from_date, to_date) -> List[WorkRecord]:
        handler = WorkRecordHandler(self._url, self._api_key)
        work_records = handler.get_work_records(from_date, to_date)

        return work_records

    def get_user_by_id(self, user_id: str) -> User:
        handler = UserHandler(self._url, self._api_key)

        user = handler.get_user_by_id(user_id)

        return user

    def get_task_by_id(self, task_id: str) -> Task:
        handler = TaskHandler(self._url, self._api_key)

        task = handler.get_task_by_id(task_id)

        return task

    def get_packages_by_project_id(self, project_id: str) -> List[Package]:
        handler = PackageHandler(self._url, self._api_key)

        packages = handler.get_packages_by_project_id(project_id)

        return packages

    def get_package_by_id(self, package_id: str) -> Package:
        handler = PackageHandler(self._url, self._api_key)

        package = handler.get_package_by_id(package_id)

        return package

    def get_project_by_id(self, project_id: str) -> Project:
        handler = ProjectHandler(self._url, self._api_key)

        project = handler.get_project_by_id(project_id)

        return project

    def get_project_structure(self, project_id: str) -> ProjectStructure:
        handler = ProjectHandler(self._url, self._api_key)

        structure = handler.get_project_structure(project_id)

        return structure

    def get_projects(self) -> List[Project]:
        handler = ProjectHandler(self._url, self._api_key)

        projects = handler.get_projects()

        return projects

    def get_contact_by_id(self, contact_id: str) -> Contact:
        handler = ContactHandler(self._url, self._api_key)

        contact = handler.get_contact_by_id(contact_id)

        return contact

    def get_company_by_id(self, company_id: str) -> Company:
        handler = CompanyHandler(self._url, self._api_key)

        company = handler.get_company_by_id(company_id)

        return company



from ctypes import Structure
from importlib.resources import Package
from typing import List, Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Package:
    id: str
    mandant_id: str
    title: str
    number: int
    change_date: str
    creator_id: str
    custom_fields: Optional[dict]
    description: Optional[str] = ''
    end_date: Optional[str] = ''
    start_date: Optional[str] = ''
    company_id: Optional[str] = ''
    company_contact_id: Optional[str] = ''
    officer_id: Optional[str] = ''
    parent_package_id: Optional[str] = ''
    planned_effort: Optional[float] = 0.0
    project_id: Optional[str] = ''
    realized_effort: Optional[float] = 0.0
    remaining_effort: Optional[float] = 0.0


class PackageHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(PackageHandler, self).__init__(url, api_key)

    def get_packages_by_project_id(self, project_id) -> List[Package]:
        path = f"/api/core/projects/{project_id}/packages"

        json_data = self.do_get(path)

        packages = Package.schema().load(json_data, many=True)

        return packages

    def get_package_by_id(self, package_id) -> Package:
        path = f"/api/core/packages/{package_id}"

        json_data = self.do_get(path)

        package = Package.schema().load(json_data)

        return package



from typing import Dict
from typing_extensions import Protocol
import json

import requests

from factro_client.core.exceptions import UnauthorizedException

class IsDataclass(Protocol):
    __dataclass_fields__: Dict

class BaseHandler(object):

    def __init__(self, url: str, api_key: str):
        self._url = url
        self._api_key = api_key


    def do_get(self, path: str, options: Dict={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        response = requests.get(request_url, headers=headers)

        self.__check_response(response)

        json_body = response.json()

        return json_body

    def do_delete(self, path: str, options: Dict={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        response = requests.delete(request_url, headers=headers)

        self.__check_response(response)

    def do_post(self, path: str, payload: IsDataclass, options: Dict={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        json_payload = json.dumps(payload)

        response = requests.post(request_url, json_payload, headers=headers)

        self.__check_response(response)

        if response.status_code == 200:
            json_body = response.json()
        else:
            json_body = {}

        return json_body

    def do_put(self, path: str, payload: IsDataclass, options: Dict={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        json_payload = json.dumps(payload)

        response = requests.put(request_url, json_payload, headers=headers)
        
        self.__check_response(response)

        if response.status_code == 200:
            json_body = response.json()
        else:
            json_body = {}

        return json_body

    def __check_response(self, response):
        if response.status_code == 401:
            raise UnauthorizedException()
        else:
            response.raise_for_status()

    def __get_auth_headers(self):
        return {'Authorization': self._api_key}

    def __get_default_headers(self):
        return {'Content-Type': 'application/json'}



from typing import List, Optional
import logging


from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .base_handler import BaseHandler

logger = logging.getLogger('factro_client')

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Project:
    id: str
    title: str
    number: int
    creator_id: str
    mandant_id: str
    is_archived: bool
    is_draft: bool
    project_state: str
    custom_fields: Optional[dict]
    description: Optional[str] = ''
    end_date: Optional[str] = ''
    start_date: Optional[str] = ''
    company_id: Optional[str] = ''
    company_contact_id: Optional[str] = ''
    officer_id: Optional[str] = ''
    planned_effort: Optional[float] = 0.0
    realized_effort: Optional[float] = 0.0
    remaining_effort: Optional[float] = 0.0
    priority: Optional[int] = 0

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ProjectStructure:
    id: str
    type: str
    children: "List[Optional[ProjectStructure]]"


class ProjectHandler(BaseHandler):

    def __init__(self, url: str, api_key: str):
        super(ProjectHandler, self).__init__(url, api_key)

    def get_project_by_id(self, project_id) -> Project:
        path = f"/api/core/projects/{project_id}"

        json_data = self.do_get(path)

        logger.debug(f"ProjectHandler.get_project_by_id({project_id} -> {json_data}")

        project = Project.schema().load(json_data)

        return project

    def get_projects(self) -> List[Project]:
        path = f"/api/core/projects"
        json_data = self.do_get(path)

        projects = Project.schema().load(json_data, many=True)

        return projects

    def get_project_structure(self, project_id: str) -> ProjectStructure:
        path = f"/api/core/projects/{project_id}/structure"

        json_data = self.do_get(path)

        structure = ProjectStructure.schema().load(json_data)

        return structure



from datetime import datetime, timedelta
import logging
import os

from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase
from typing import List, Optional

from cachetools import TTLCache, cached


from ..core.client import Client as CoreClient
from ..core.client import Project, Package, Task

logger = logging.getLogger('factro_client')


CACHE_MAX_SIZE = os.environ.get('FACTRO_CLIENT_CACHE_MAX_SIZE', 100)
CACHE_TTL = os.environ.get('FACTRO_CLIENT_CACHE_TTL', 2 * 60 * 60)  # 2 hours

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ProjectIntro:
    id: str
    number: int
    link: str
    title: str
    project_state: str
    start_date: str
    end_date: str
    custom_fields: Optional[dict]
    billing_cycle: str
    offer_number: str
    order_number: str
    description: Optional[str] = ''

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class EffortEstimationTask:
    id: str
    number: str
    link: str
    title: str
    task_state: str
    planned_effort: float
    realized_effort: float
    remaining_effort: float
    custom_fields: Optional[dict]
    role: Optional[str] = ''
    description: Optional[str] = ''

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class EffortEstimationPackage:
    id: str
    number: int
    link: str
    title: str
    custom_fields: Optional[dict]
    tasks: List[EffortEstimationTask]
    description: Optional[str] = ''
    realized_effort: Optional[float] = 0.0
    remaining_effort: Optional[float] = 0.0

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class EffortEstimationProjekt:
    id: str
    number: int
    link: str
    title: str
    project_state: str
    custom_fields: Optional[dict]
    billing_cycle: str
    offer_number: str
    order_number: str
    packages: List[EffortEstimationPackage]
    description: Optional[str] = ''
    planned_effort: Optional[float] = 0.0
    realized_effort: Optional[float] = 0.0
    remaining_effort: Optional[float] = 0.0

class Client:

    def __init__(self, url: str, api_key: str):
        self._core_client = CoreClient(url, api_key)

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_project_by_id(self, project_id: str) -> Project:
        project = self._core_client.get_project_by_id(project_id)
        return project
        
    @cached(cache=TTLCache(maxsize=100, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_package_by_id(self, package_id: str) -> Package:
        package = self._core_client.get_package_by_id(package_id)
        return package

    @cached(cache=TTLCache(maxsize=CACHE_MAX_SIZE, ttl=timedelta(seconds=CACHE_TTL), timer=datetime.now))
    def _query_task_by_id(self, task_id: str) -> Task:
        task = self._core_client.get_task_by_id(task_id)
        return task

    def _map_project_to_effort_estimation_projekt(self, project: Project) -> EffortEstimationProjekt:
        effort_estimation_projekt = EffortEstimationProjekt(
            id = project.id,
            number = project.number,
            # https://cloud.factro.com/5681c8b0/?p=task&pi=563227c4-5259-4953-bdea-286f41a8790c
            link = f"{self._core_client._url}/?p=project&pi={project.id}",
            title = project.title,
            project_state = project.project_state,
            custom_fields = project.custom_fields,
            billing_cycle = project.custom_fields.get('abrechnungsart', ''),
            offer_number = project.custom_fields.get('angebotsnummer', ''),
            order_number = project.custom_fields.get('bestellnummer', ''),
            packages = [],
            description = project.description,
            planned_effort = project.planned_effort,
            realized_effort = project.realized_effort,
            remaining_effort = project.remaining_effort
        )

        return effort_estimation_projekt

    def _handle_task(self, structure_task) -> EffortEstimationTask:
        task = self._query_task_by_id(structure_task.id)

        effort_estimation_task = EffortEstimationTask(
            id = task.id,
            number = task.number,
            # https://cloud.factro.com/5681c8b0/?p=task&pi=563227c4-5259-4953-bdea-286f41a8790c
            link = f"{self._core_client._url}/?p=task&pi={task.id}",
            title = task.title,
            task_state = task.task_state,
            planned_effort = task.planned_effort,
            realized_effort = task.realized_effort,
            remaining_effort = task.remaining_effort,
            custom_fields = task.custom_fields,
            role = task.custom_fields.get('rolle', ''),
            description = task.description,
        )

        return effort_estimation_task

    def _find_tasks(self, structure_childs) -> List[EffortEstimationTask]:
        effort_estimation_tasks = []

        for child in structure_childs:
            if child.type == 'task':
                task = self._handle_task(child)
                effort_estimation_tasks.append(task)
            elif child.type == 'package':
                tasks = self._find_tasks(child.children)
                effort_estimation_tasks.extend(tasks)

        return effort_estimation_tasks

    def _handle_package(self, structure_package) -> EffortEstimationPackage:

        package = self._query_package_by_id(structure_package.id)
        tasks = self._find_tasks(structure_package.children)

        effort_estimation_package = EffortEstimationPackage(
            id = package.id,
            number = package.number,
            # https://cloud.factro.com/5681c8b0/?p=task&pi=563227c4-5259-4953-bdea-286f41a8790c
            link = f"{self._core_client._url}/?p=package&pi={package.id}",
            title = package.title,
            custom_fields = package.custom_fields,
            tasks = tasks,
            description = package.description,
            realized_effort = package.realized_effort,
            remaining_effort = package.remaining_effort
        )

        return effort_estimation_package
        

    def query_effort_estimation_projekt(self, project_id: str) -> EffortEstimationProjekt:

        effort_estimation_projekt = None

        structure = self._core_client.get_project_structure(project_id)
        project = self._query_project_by_id(structure.id)
        effort_estimation_projekt = self._map_project_to_effort_estimation_projekt(project)

        for child in structure.children:
            if child.type == 'task':
                logger.warning(f"Cannot handle a task ('{child.id}', '{child.type}') within a project {structure.id}")
            elif child.type == 'package':
                effort_estimation_package = self._handle_package(child)
                effort_estimation_projekt.packages.append(effort_estimation_package)

        return effort_estimation_projekt

    def query_projects(self) -> List[ProjectIntro]:
        
        def _map_project(project: Project) -> ProjectIntro:
            project_intro = ProjectIntro(
                id = project.id,
                number = project.number,
                # https://cloud.factro.com/5681c8b0/?p=task&pi=563227c4-5259-4953-bdea-286f41a8790c
                link = f"{self._core_client._url}/?p=project&pi={project.id}",
                title = project.title,
                project_state = project.project_state,
                start_date = project.start_date,
                end_date = project.end_date,
                custom_fields = project.custom_fields,
                billing_cycle = project.custom_fields.get('abrechnungsart', ''),
                offer_number = project.custom_fields.get('angebotsnummer', ''),
                order_number = project.custom_fields.get('bestellnummer', ''),
                description = project.description
            )

            return project_intro
        
        project_intros = []
        projects = self._core_client.get_projects()

        for project in projects:
            project_intro = _map_project(project)
            project_intros.append(project_intro)

        return project_intros



import setuptools

try:
    from pip.req import parse_requirements
except ImportError:  # pip >= 10.0.0
    from pip._internal.req import parse_requirements

# TODO: mm - twine benutzen https://pypi.org/project/twine/
#
# Kurzanleitung
#    https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56


parsed_reqs = parse_requirements('requirements.txt', session='hack')
installed_reqs = [str(ir.requirement) for ir in parsed_reqs]

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='factro_client',
    version=setuptools.sic('0.4.1-alpha.1'),
    author="5Minds IT-Solutions GmbH & Co. KG",
    author_email="developers@5minds.de",
    description="A core api client for project management tool 'factro'.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="client factro 5minds processcube",
    url="https://github.com/5minds/factro_client",
    packages=setuptools.find_packages(),
    install_requires=installed_reqs,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)